@@ -13,6 +13,7 @@
 
 #include <glib/gstdio.h>
 #include <glib/gi18n.h>
+#include <gio/gio.h>
 
 #include <string.h>
 
@@ -20,22 +21,33 @@
     #include "config.h"
 #endif
 
+#if HAVE_UNISTD_H
+    #include <unistd.h>
+#endif
+
 #ifdef HAVE_HILDON_2_2
     #include <hildon/hildon.h>
 #endif
 
+#if !GTK_CHECK_VERSION (2, 18, 0)
+    #define gtk_widget_get_has_window(wdgt) !GTK_WIDGET_NO_WINDOW (wdgt)
+    #define gtk_widget_get_allocation (wdgt, alloc) *alloc = wdgt->allocation
+#endif
+
 static void
 proxy_toggle_button_toggled_cb (GtkToggleButton* button,
                                 GObject*         object)
 {
     gboolean toggled;
+    const gchar* property;
+
     #ifdef HAVE_HILDON_2_2
     if (HILDON_IS_CHECK_BUTTON (button))
         toggled = hildon_check_button_get_active (HILDON_CHECK_BUTTON (button));
     #else
     toggled = gtk_toggle_button_get_active (button);
     #endif
-    const gchar* property = g_object_get_data (G_OBJECT (button), "property");
+    property = g_object_get_data (G_OBJECT (button), "property");
     g_object_set (object, property, toggled, NULL);
 }
 
@@ -324,6 +336,95 @@ katze_app_info_get_all_for_category (const gchar* category)
     return apps;
 }
 
+#if !GLIB_CHECK_VERSION (2, 20, 0)
+/* Icon tokenization, for Glib < 2.20 */
+static gboolean
+g_icon_to_string_tokenized (GIcon   *icon,
+                            GString *s)
+{
+  GPtrArray *tokens;
+  guint i;
+
+  if (G_IS_THEMED_ICON (icon))
+    {
+      guint n;
+      const char * const *names;
+      tokens = g_ptr_array_new ();
+      g_object_get (icon, "names", &names, NULL);
+      for (n = 0; names[n] != NULL; n++)
+        g_ptr_array_add (tokens, g_strdup (names[n]));
+    }
+  else if (G_IS_FILE_ICON (icon))
+    {
+      tokens = g_ptr_array_new ();
+      g_ptr_array_add (tokens,
+                       g_file_get_uri (g_file_icon_get_file (G_FILE_ICON (icon))));
+    }
+  else
+    return FALSE;
+
+  g_string_append (s, g_type_name_from_instance ((GTypeInstance *)icon));
+
+  for (i = 0; i < tokens->len; i++)
+    {
+      char *token = g_ptr_array_index (tokens, i);
+
+      g_string_append_c (s, ' ');
+      g_string_append_uri_escaped (s, token, "!$&'()*+,;=:@/", TRUE);
+
+      g_free (token);
+    }
+
+  g_ptr_array_free (tokens, TRUE);
+
+  return TRUE;
+}
+
+/* GIcon serialization, for Glib < 2.20 */
+static gchar *
+g_icon_to_string (GIcon *icon)
+{
+  gchar *ret = NULL;
+
+  g_return_val_if_fail (G_IS_ICON (icon), NULL);
+
+  if (G_IS_FILE_ICON (icon))
+    {
+      GFile *file = g_file_icon_get_file (G_FILE_ICON (icon));
+      if (g_file_is_native (file))
+        {
+          ret = g_file_get_path (file);
+          if (!g_utf8_validate (ret, -1, NULL))
+            {
+              g_free (ret);
+              ret = NULL;
+            }
+        }
+      else
+        ret = g_file_get_uri (file);
+    }
+  else if (G_IS_THEMED_ICON (icon))
+    {
+      const char * const *names;
+      g_object_get (icon, "names", &names, NULL);
+      if (names && names[0] && names[0][0] != '.' &&
+          g_utf8_validate (names[0], -1, NULL) && names[1] == NULL)
+        ret = g_strdup (names[0]);
+    }
+
+  if (ret == NULL)
+    {
+      GString *s = g_string_new (". ");
+      if (g_icon_to_string_tokenized (icon, s))
+        ret = g_string_free (s, FALSE);
+      else
+        g_string_free (s, TRUE);
+    }
+
+  return ret;
+}
+#endif
+
 /**
  * katze_property_proxy:
  * @object: a #GObject
@@ -496,6 +597,9 @@ katze_property_proxy (gpointer     object,
         pango_context_list_families (context, &families, &n_families);
         if (!string)
             string = g_strdup (G_PARAM_SPEC_STRING (pspec)->default_value);
+        /* 'sans' and 'sans-serif' are presumably the same */
+        if (!g_strcmp0 (string, "sans-serif"))
+            katze_assign (string, g_strdup ("sans"));
         if (string)
         {
             gint j = 0;
@@ -523,12 +627,20 @@ katze_property_proxy (gpointer     object,
         GtkComboBox* combo;
         GList* apps;
         const gchar* app_type = &hint[12];
+        GtkSettings* settings;
+        gint icon_width = 16;
 
-        model = gtk_list_store_new (3, G_TYPE_APP_INFO, G_TYPE_STRING, G_TYPE_STRING);
+        settings = gtk_settings_get_for_screen (gdk_screen_get_default ());
+        gtk_icon_size_lookup_for_settings (settings, GTK_ICON_SIZE_MENU,
+                                           &icon_width, NULL);
+
+        model = gtk_list_store_new (4, G_TYPE_APP_INFO, G_TYPE_STRING,
+                                       G_TYPE_STRING, G_TYPE_INT);
         widget = gtk_combo_box_new_with_model (GTK_TREE_MODEL (model));
         renderer = gtk_cell_renderer_pixbuf_new ();
         gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (widget), renderer, FALSE);
         gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (widget), renderer, "icon-name", 1);
+        gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (widget), renderer, "width", 3);
         renderer = gtk_cell_renderer_text_new ();
         gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (widget), renderer, TRUE);
         gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (widget), renderer, "text", 2);
@@ -548,17 +660,21 @@ katze_property_proxy (gpointer     object,
             GAppInfo* info;
 
             gtk_list_store_insert_with_values (model, &iter_none, 0,
-                0, NULL, 1, NULL, 2, _("None"), -1);
+                0, NULL, 1, NULL, 2, _("None"), 3, icon_width, -1);
 
             while ((info = g_list_nth_data (apps, i++)))
             {
                 const gchar* name = g_app_info_get_name (info);
                 GIcon* icon = g_app_info_get_icon (info);
-                gchar* icon_name = g_icon_to_string (icon);
+                gchar* icon_name;
                 GtkTreeIter iter;
 
+                if (!g_app_info_should_show (info))
+                    continue;
+
+                icon_name = icon ? g_icon_to_string (icon) : NULL;
                 gtk_list_store_insert_with_values (model, &iter, G_MAXINT,
-                    0, info, 1, icon_name, 2, name, -1);
+                    0, info, 1, icon_name, 2, name, 3, icon_width, -1);
                 if (string && !strcmp (katze_app_info_get_commandline (info), string))
                     gtk_combo_box_set_active_iter (combo, &iter);
 
@@ -568,7 +684,7 @@ katze_property_proxy (gpointer     object,
             info = g_app_info_create_from_commandline ("",
                 "", G_APP_INFO_CREATE_NONE, NULL);
             gtk_list_store_insert_with_values (model, NULL, G_MAXINT,
-                0, info, 1, NULL, 2, _("Custom..."), -1);
+                0, info, 1, NULL, 2, _("Custom..."), 3, icon_width, -1);
             g_object_unref (info);
 
             if (gtk_combo_box_get_active (combo) == -1)
@@ -644,6 +760,10 @@ katze_property_proxy (gpointer     object,
         widget = gtk_spin_button_new_with_range (
             G_PARAM_SPEC_INT (pspec)->minimum,
             G_PARAM_SPEC_INT (pspec)->maximum, 1);
+        #if HAVE_HILDON
+        hildon_gtk_entry_set_input_mode (GTK_ENTRY (widget),
+                                         HILDON_GTK_INPUT_MODE_NUMERIC);
+        #endif
         /* Keep it narrow, 5 digits are usually fine */
         gtk_entry_set_width_chars (GTK_ENTRY (widget), 5);
         gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), value);
@@ -778,7 +898,9 @@ katze_property_label (gpointer     object,
 
     nick = g_param_spec_get_nick (pspec);
     widget = gtk_label_new (nick);
+    #if GTK_CHECK_VERSION (2, 12, 0)
     gtk_widget_set_tooltip_text (widget, g_param_spec_get_blurb (pspec));
+    #endif
     gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
 
     return widget;
@@ -799,21 +921,24 @@ katze_widget_popup_position_menu (GtkMenu*  menu,
 {
     gint wx, wy;
     gint menu_width;
+    GtkAllocation allocation;
     GtkRequisition menu_req;
     GtkRequisition widget_req;
     KatzePopupInfo* info = user_data;
     GtkWidget* widget = info->widget;
     gint widget_height;
 
+    gtk_widget_get_allocation (widget, &allocation);
+
     /* Retrieve size and position of both widget and menu */
-    if (GTK_WIDGET_NO_WINDOW (widget))
+    if (!gtk_widget_get_has_window (widget))
     {
-        gdk_window_get_position (widget->window, &wx, &wy);
-        wx += widget->allocation.x;
-        wy += widget->allocation.y;
+        gdk_window_get_position (gtk_widget_get_window (widget), &wx, &wy);
+        wx += allocation.x;
+        wy += allocation.y;
     }
     else
-        gdk_window_get_origin (widget->window, &wx, &wy);
+        gdk_window_get_origin (gtk_widget_get_window (widget), &wx, &wy);
     gtk_widget_size_request (GTK_WIDGET (menu), &menu_req);
     gtk_widget_size_request (widget, &widget_req);
     menu_width = menu_req.width;
@@ -824,7 +949,7 @@ katze_widget_popup_position_menu (GtkMenu*  menu,
         ; /* Do nothing? */
     else if (info->position == KATZE_MENU_POSITION_RIGHT)
     {
-        *x = wx + widget->allocation.width - menu_width;
+        *x = wx + allocation.width - menu_width;
         *y = wy + widget_height;
     } else if (info->position == KATZE_MENU_POSITION_LEFT)
     {
@@ -1233,7 +1358,8 @@ katze_mkdir_with_parents (const gchar* pathname,
 {
   gchar* fn, *p;
 
-  if (g_file_test (pathname, G_FILE_TEST_EXISTS))
+  /* Use g_access instead of g_file_test for better performance */
+  if (g_access (pathname, F_OK) == 0)
       return 0;
 
   fn = g_strdup (pathname);
@@ -1253,7 +1379,7 @@ katze_mkdir_with_parents (const gchar* pathname,
       else
           *p = '\0';
 
-      if (!g_file_test (fn, G_FILE_TEST_EXISTS))
+      if (g_access (fn, F_OK) != 0)
       {
           if (g_mkdir (fn, mode) == -1)
           {
@@ -1312,3 +1438,155 @@ katze_widget_has_touchscreen_mode (GtkWidget* widget)
         return enabled;
     }
 }
+
+/**
+ * katze_load_cached_icon:
+ * @uri: an URI string
+ * @widget: a #GtkWidget, or %NULL
+ *
+ * Loads a cached icon for the specified @uri. If there is no
+ * icon and @widget is specified, a default will be returned.
+ *
+ * Returns: a #GdkPixbuf, or %NULL
+ *
+ * Since: 0.2.2
+ */
+GdkPixbuf*
+katze_load_cached_icon (const gchar* uri,
+                        GtkWidget*   widget)
+{
+    GdkPixbuf* icon = NULL;
+
+    if (g_str_has_prefix (uri, "http://"))
+    {
+        guint i;
+        gchar* icon_uri;
+        gchar* checksum;
+        gchar* ext;
+        gchar* filename;
+        gchar* path;
+
+        i = 8;
+        while (uri[i] != '\0' && uri[i] != '/')
+            i++;
+        if (uri[i] == '/')
+        {
+            gchar* ticon_uri = g_strdup (uri);
+            ticon_uri[i] = '\0';
+            icon_uri = g_strdup_printf ("%s/favicon.ico", ticon_uri);
+            g_free (ticon_uri);
+        }
+        else
+            icon_uri = g_strdup_printf ("%s/favicon.ico", uri);
+
+        checksum = g_compute_checksum_for_string (G_CHECKSUM_MD5, icon_uri, -1);
+        ext = g_strrstr (icon_uri, ".");
+        filename = g_strdup_printf ("%s%s", checksum, ext ? ext : "");
+        g_free (icon_uri);
+        g_free (checksum);
+        path = g_build_filename (g_get_user_cache_dir (), PACKAGE_NAME,
+                                 "icons", filename, NULL);
+        if ((icon = gdk_pixbuf_new_from_file_at_size (path, 16, 16, NULL)))
+        {
+            g_free (path);
+            return icon;
+        }
+        g_free (path);
+    }
+
+    return icon || !widget ? icon : gtk_widget_render_icon (widget,
+        GTK_STOCK_FILE, GTK_ICON_SIZE_MENU, NULL);
+}
+
+/**
+ * katze_collfold:
+ * @str: a non-NULL UTF-8 string
+ *
+ * Computes a string without case and decomposited so
+ * it can be used for comparison.
+ *
+ * Return value: a normalized string
+ *
+ * Since: 0.2.3
+ **/
+gchar*
+katze_collfold (const gchar* str)
+{
+    GString* result = g_string_new (NULL);
+    const gchar* p = str;
+
+    while (*p)
+    {
+        gunichar ch = g_unichar_tolower (g_utf8_get_char (p));
+        gsize len;
+        gunichar* sch = g_unicode_canonical_decomposition (ch, &len);
+        guint i = 0;
+        while (i < len)
+            g_string_append_unichar (result, sch[i++]);
+
+        p = g_utf8_next_char (p);
+    }
+
+    return g_string_free (result, FALSE);
+}
+
+/**
+ * katze_utf8_stristr:
+ * @haystack: a non-NULL UTF-8 string
+ * @needle: a normalized non-NULL UTF-8 string
+ *
+ * Determines whether @needle is in @haystack, disregarding
+ * differences in case.
+ *
+ * Return value: %TRUE if @needle is found in @haystack
+ *
+ * Since: 0.2.3
+ **/
+gboolean
+katze_utf8_stristr (const gchar* haystack,
+                    const gchar* needle)
+{
+    #if 0 /* 0,000159 seconds */
+    /* Too slow for use in completion */
+    gchar* nhaystack = g_utf8_normalize (haystack, -1, G_NORMALIZE_DEFAULT);
+    const gchar *p = nhaystack;
+    gsize len = strlen (needle);
+    gsize i;
+
+    while (*p)
+    {
+        for (i = 0; i < len; i++)
+            if (g_unichar_tolower (g_utf8_get_char (p + i))
+             != g_unichar_tolower (g_utf8_get_char (needle + i)))
+                goto next;
+
+        g_free (nhaystack);
+        return TRUE;
+
+        next:
+            p = g_utf8_next_char (p);
+    }
+
+    g_free (nhaystack);
+    return FALSE;
+    #else /* 0,000044 seconds */
+    /* No unicode matching */
+    const gchar *p = haystack;
+    gsize len = strlen (needle);
+    gsize i;
+
+    while (*p)
+    {
+        for (i = 0; i < len; i++)
+            if (g_ascii_tolower (p[i]) != g_ascii_tolower (needle[i]))
+                goto next;
+
+        return TRUE;
+
+        next:
+            p++;
+    }
+
+    return FALSE;
+    #endif
+}
