@@ -22,6 +22,11 @@
 
 #if HAVE_HILDON
     #include <libosso.h>
+    #ifdef HAVE_HILDON_2_2
+        #include <dbus/dbus.h>
+        #include <mce/mode-names.h>
+        #include <mce/dbus-names.h>
+    #endif
     typedef osso_context_t* MidoriAppInstance;
     #define MidoriAppInstanceNull NULL
 #elif HAVE_UNIQUE
@@ -35,20 +40,8 @@
     #include "socket.h"
 #endif
 
-#if !HAVE_HILDON
-typedef struct _NotifyNotification NotifyNotification;
-
-typedef struct
-{
-    gboolean            (*init)               (const gchar* app_name);
-    void                (*uninit)             (void);
-    NotifyNotification* (*notification_new)   (const gchar* summary,
-                                               const gchar* body,
-                                               const gchar* icon,
-                                               GtkWidget*   attach);
-    gboolean            (*notification_show)  (NotifyNotification* notification,
-                                               GError**            error);
-} LibNotifyFuncs;
+#if HAVE_LIBNOTIFY
+    #include <libnotify/notify.h>
 #endif
 
 struct _MidoriApp
@@ -69,11 +62,8 @@ struct _MidoriApp
 
     MidoriAppInstance instance;
 
-    #if !HAVE_HILDON
-    /* libnotify handling */
-    gchar*         program_notify_send;
-    GModule*       libnotify_module;
-    LibNotifyFuncs libnotify_funcs;
+    #if !HAVE_HILDON || !HAVE_LIBNOTIFY
+    gchar* program_notify_send;
     #endif
 };
 
@@ -123,9 +113,6 @@ static guint signals[LAST_SIGNAL];
 static void
 midori_app_finalize (GObject* object);
 
-static void
-midori_app_init_libnotify (MidoriApp* app);
-
 static void
 midori_app_set_property (GObject*      object,
                          guint         prop_id,
@@ -219,8 +206,10 @@ _midori_app_add_browser (MidoriApp*     app,
 
     app->browser = browser;
     #if HAVE_UNIQUE
+    /* We *do not* let unique watch windows because that includes
+        bringing windows in the foreground, even from other workspaces.
     if (app->instance)
-        unique_app_watch_window (app->instance, GTK_WINDOW (browser));
+        unique_app_watch_window (app->instance, GTK_WINDOW (browser)); */
     #endif
 }
 
@@ -404,6 +393,15 @@ midori_app_class_init (MidoriAppClass* class)
                                      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 }
 
+static void
+midori_app_raise_window (GtkWindow* window,
+                         GdkScreen* screen)
+{
+    gtk_window_set_screen (window, screen);
+    gtk_window_present (window);
+    gtk_window_deiconify (window);
+}
+
 static gboolean
 midori_app_command_received (MidoriApp*   app,
                              const gchar* command,
@@ -423,8 +421,7 @@ midori_app_command_received (MidoriApp*   app,
         if (!app->browser)
             return FALSE;
 
-        gtk_window_set_screen (GTK_WINDOW (app->browser), screen);
-        gtk_window_present (GTK_WINDOW (app->browser));
+        midori_app_raise_window (GTK_WINDOW (app->browser), screen);
         return TRUE;
     }
     else if (g_str_equal (command, "new"))
@@ -434,8 +431,8 @@ midori_app_command_received (MidoriApp*   app,
         /* FIXME: Should open the homepage according to settings */
         midori_browser_add_uri (browser, "");
         midori_browser_activate_action (browser, "Location");
-        gtk_window_set_screen (GTK_WINDOW (app->browser), screen);
         gtk_widget_show (GTK_WIDGET (browser));
+        midori_app_raise_window (GTK_WINDOW (browser), screen);
         return TRUE;
     }
     else if (g_str_equal (command, "open"))
@@ -443,8 +440,6 @@ midori_app_command_received (MidoriApp*   app,
         if (!app->browser)
             return FALSE;
 
-        gtk_window_set_screen (GTK_WINDOW (app->browser), screen);
-        gtk_window_present (GTK_WINDOW (app->browser));
         if (!uris)
             return FALSE;
         else
@@ -459,15 +454,19 @@ midori_app_command_received (MidoriApp*   app,
             {
                 browser = midori_app_create_browser (app);
                 midori_app_add_browser (app, browser);
-                gtk_window_set_screen (GTK_WINDOW (app->browser), screen);
                 gtk_widget_show (GTK_WIDGET (browser));
             }
             else
                 browser = app->browser;
+
+            midori_app_raise_window (GTK_WINDOW (browser), screen);
+
             first = (open_external_pages_in == MIDORI_NEW_PAGE_CURRENT);
             while (*uris)
             {
-                gchar* fixed_uri = sokoke_magic_uri (*uris, NULL);
+                gchar* fixed_uri = sokoke_magic_uri (*uris);
+                if (!fixed_uri)
+                    fixed_uri = g_strdup (*uris);
                 if (first)
                 {
                     midori_browser_set_current_uri (browser, fixed_uri);
@@ -484,9 +483,15 @@ midori_app_command_received (MidoriApp*   app,
     }
     else if (g_str_equal (command, "command"))
     {
+        guint i = 0;
+
         if (!uris || !app->browser)
             return FALSE;
-        midori_browser_activate_action (app->browser, *uris);
+        while (uris[i] != NULL)
+        {
+            midori_browser_activate_action (app->browser, uris[i]);
+            i++;
+        }
         return TRUE;
     }
 
@@ -612,8 +617,6 @@ midori_app_io_channel_watch_cb (GIOChannel*  channel,
         }
     }
 
-    gtk_window_present (GTK_WINDOW (app->browser));
-
     fd_close (sock);
 
     return TRUE;
@@ -642,6 +645,12 @@ midori_app_create_instance (MidoriApp*   app,
         osso_deinitialize (instance);
         return NULL;
     }
+
+    #ifdef HAVE_HILDON_2_2
+    if (OSSO_OK == osso_rpc_run_system (instance, MCE_SERVICE, MCE_REQUEST_PATH,
+        MCE_REQUEST_IF, MCE_ACCELEROMETER_ENABLE_REQ, NULL, DBUS_TYPE_INVALID))
+        /* Accelerometer enabled */;
+    #endif
     #else
     GdkDisplay* display;
     gchar* display_name;
@@ -661,7 +670,8 @@ midori_app_create_instance (MidoriApp*   app,
     display_name = g_strdup (gdk_display_get_name (display));
     n = strlen (display_name);
     for (i = 0; i < n; i++)
-        if (display_name[i] == ':' || display_name[i] == '.')
+        if (display_name[i] == ':' || display_name[i] == '.'
+         || display_name[i] == '\\')
             display_name[i] = '_';
     instance_name = g_strdup_printf ("de.twotoasts.%s_%s", name, display_name);
 
@@ -674,9 +684,12 @@ midori_app_create_instance (MidoriApp*   app,
     instance = socket_init (instance_name, sokoke_set_config_dir (NULL), &exists);
     g_object_set_data (G_OBJECT (app), "sock-exists",
         exists ? (gpointer)0xdeadbeef : NULL);
-    channel = g_io_channel_unix_new (instance);
-    g_io_add_watch (channel, G_IO_IN | G_IO_PRI | G_IO_ERR,
-        (GIOFunc)midori_app_io_channel_watch_cb, app);
+    if (instance != MidoriAppInstanceNull)
+    {
+        channel = g_io_channel_unix_new (instance);
+        g_io_add_watch (channel, G_IO_IN | G_IO_PRI | G_IO_ERR,
+            (GIOFunc)midori_app_io_channel_watch_cb, app);
+    }
     #endif
 
     g_free (instance_name);
@@ -701,7 +714,12 @@ midori_app_init (MidoriApp* app)
 
     app->instance = MidoriAppInstanceNull;
 
-    midori_app_init_libnotify (app);
+    #if HAVE_LIBNOTIFY
+    notify_init ("midori");
+    #else
+    app->program_notify_send = g_find_program_in_path ("notify-send");
+    #endif
+
 }
 
 static void
@@ -729,13 +747,11 @@ midori_app_finalize (GObject* object)
     sock_cleanup ();
     #endif
 
-    #if !HAVE_HILDON
-    if (app->libnotify_module)
-    {
-        app->libnotify_funcs.uninit ();
-        g_module_close (app->libnotify_module);
-    }
-    katze_assign (app->program_notify_send, NULL);
+    #if HAVE_LIBNOTIFY
+    if (notify_is_initted ())
+        notify_uninit ();
+    #else
+        katze_assign (app->program_notify_send, NULL);
     #endif
 
     G_OBJECT_CLASS (midori_app_parent_class)->finalize (object);
@@ -1125,41 +1141,6 @@ midori_app_quit (MidoriApp* app)
     g_signal_emit (app, signals[QUIT], 0);
 }
 
-static void
-midori_app_init_libnotify (MidoriApp* app)
-{
-    #if !HAVE_HILDON
-    gint i;
-    const gchar* sonames[] = { "libnotify.so", "libnotify.so.1", NULL };
-
-    for (i = 0; sonames[i] != NULL && app->libnotify_module == NULL; i++ )
-    {
-        app->libnotify_module = g_module_open (sonames[i], G_MODULE_BIND_LOCAL);
-    }
-
-    if (app->libnotify_module != NULL)
-    {
-        g_module_symbol (app->libnotify_module, "notify_init",
-            (void*) &(app->libnotify_funcs.init));
-        g_module_symbol (app->libnotify_module, "notify_uninit",
-            (void*) &(app->libnotify_funcs.uninit));
-        g_module_symbol (app->libnotify_module, "notify_notification_new",
-            (void*) &(app->libnotify_funcs.notification_new));
-        g_module_symbol (app->libnotify_module, "notify_notification_show",
-            (void*) &(app->libnotify_funcs.notification_show));
-
-        /* init libnotify */
-        if (!app->libnotify_funcs.init || !app->libnotify_funcs.init ("midori"))
-        {
-             g_module_close (app->libnotify_module);
-             app->libnotify_module = NULL;
-        }
-    }
-
-    app->program_notify_send = g_find_program_in_path ("notify-send");
-    #endif
-}
-
 /**
  * midori_app_send_notification:
  * @app: a #MidoriApp
@@ -1178,23 +1159,25 @@ midori_app_send_notification (MidoriApp*   app,
                               const gchar* title,
                               const gchar* message)
 {
+    g_return_if_fail (MIDORI_IS_APP (app));
+    g_return_if_fail (title);
+
     #if HAVE_HILDON
     hildon_banner_show_information_with_markup (GTK_WIDGET (app->browser),
                                                 "midori", message);
     #else
     gboolean sent = FALSE;
 
-    g_return_if_fail (MIDORI_IS_APP (app));
-    g_return_if_fail (title);
-
-    if (app->libnotify_module)
+    #if HAVE_LIBNOTIFY
+    if (notify_is_initted ())
     {
-        NotifyNotification* n;
+        NotifyNotification* note;
 
-        n = app->libnotify_funcs.notification_new (title, message, "midori", NULL);
-        sent = app->libnotify_funcs.notification_show (n, NULL);
-        g_object_unref (n);
+        note = notify_notification_new (title, message, "midori", NULL);
+        sent = notify_notification_show (note, NULL);
+        g_object_unref (note);
     }
+    #else
     /* Fall back to the command line program "notify-send" */
     if (!sent && app->program_notify_send)
     {
@@ -1210,4 +1193,5 @@ midori_app_send_notification (MidoriApp*   app,
         g_free (command);
     }
     #endif
+    #endif
 }
