@@ -13,7 +13,6 @@
 
 #include "midori-view.h"
 
-#include "compat.h"
 #include "marshal.h"
 #include "sokoke.h"
 
@@ -33,6 +32,7 @@ struct _MidoriPanel
     GtkWidget* toolbar;
     GtkToolItem* button_align;
     GtkToolItem* button_detach;
+    GtkToolItem* button_controls;
     GtkWidget* toolbar_label;
     GtkWidget* frame;
     GtkWidget* toolbook;
@@ -43,6 +43,7 @@ struct _MidoriPanel
     gboolean show_titles;
     gboolean show_controls;
     gboolean right_aligned;
+    gboolean open_panels_in_windows;
 };
 
 struct _MidoriPanelClass
@@ -67,6 +68,7 @@ enum
     PROP_SHOW_TITLES,
     PROP_SHOW_CONTROLS,
     PROP_RIGHT_ALIGNED,
+    PROP_OPEN_PANELS_IN_WINDOWS,
 };
 
 enum {
@@ -195,7 +197,7 @@ midori_panel_class_init (MidoriPanelClass* class)
      *
      * Whether to show panel titles.
      *
-     * Since: 0.1.9
+     * Deprecated: 0.2.3
      */
     g_object_class_install_property (gobject_class,
                                      PROP_SHOW_TITLES,
@@ -237,6 +239,22 @@ midori_panel_class_init (MidoriPanelClass* class)
                                      "Whether the panel is aligned to the right",
                                      FALSE,
                                      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+    /**
+     * MidoriWebSettings:open-panels-in-windows:
+     *
+     * Whether to open panels in separate windows.
+     *
+     * Since: 0.2.2
+     */
+    g_object_class_install_property (gobject_class,
+                                     PROP_OPEN_PANELS_IN_WINDOWS,
+                                     g_param_spec_boolean (
+                                     "open-panels-in-windows",
+                                     "Open panels in windows",
+        "Whether to open panels in standalone windows by default",
+                                     FALSE,
+                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 }
 
 static void
@@ -266,6 +284,7 @@ midori_panel_detached_window_delete_event_cb (GtkWidget*   window,
     GtkWidget* scrolled = g_object_get_data (G_OBJECT (window), "scrolled");
     GtkWidget* toolbar = g_object_get_data (G_OBJECT (scrolled), "panel-toolbar");
     GtkWidget* menuitem = g_object_get_data (G_OBJECT (scrolled), "panel-menuitem");
+    GtkWidget* viewable = _midori_panel_child_for_scrolled (panel, scrolled);
     GtkToolItem* toolitem;
     gint n;
 
@@ -277,13 +296,9 @@ midori_panel_detached_window_delete_event_cb (GtkWidget*   window,
     gtk_container_remove (GTK_CONTAINER (vbox), scrolled);
     n = gtk_notebook_append_page (GTK_NOTEBOOK (panel->notebook), scrolled, NULL);
     g_object_unref (scrolled);
-    toolitem = midori_panel_construct_tool_item (panel,
-        MIDORI_VIEWABLE (_midori_panel_child_for_scrolled (panel, scrolled)));
+    toolitem = midori_panel_construct_tool_item (panel, MIDORI_VIEWABLE (viewable));
     if (menuitem)
-    {
-        gtk_widget_show (menuitem);
         g_object_set_data (G_OBJECT (menuitem), "toolitem", toolitem);
-    }
     midori_panel_set_current_page (panel, n);
     return FALSE;
 }
@@ -298,13 +313,9 @@ midori_panel_widget_destroy_cb (GtkWidget* viewable,
 }
 
 static void
-midori_panel_button_detach_clicked_cb (GtkWidget*   toolbutton,
-                                       MidoriPanel* panel)
+midori_panel_detach_page (MidoriPanel* panel,
+                          gint         n)
 {
-    /* FIXME: What happens when the browser is destroyed? */
-    /* FIXME: What about multiple browsers? */
-    /* FIXME: Should we remember if the child was detached? */
-    gint n = midori_panel_get_current_page (panel);
     GtkToolItem* toolitem = gtk_toolbar_get_nth_item (
         GTK_TOOLBAR (panel->toolbar), n);
     const gchar* title = gtk_tool_button_get_label (GTK_TOOL_BUTTON (toolitem));
@@ -312,7 +323,6 @@ midori_panel_button_detach_clicked_cb (GtkWidget*   toolbutton,
         GTK_NOTEBOOK (panel->toolbook), n);
     GtkWidget* scrolled = gtk_notebook_get_nth_page (
         GTK_NOTEBOOK (panel->notebook), n);
-    GtkWidget* menuitem = g_object_get_data (G_OBJECT (scrolled), "panel-menuitem");
     GtkWidget* vbox = gtk_vbox_new (FALSE, 0);
     #if HAVE_HILDON
     GtkWidget* window = hildon_window_new ();
@@ -328,8 +338,6 @@ midori_panel_button_detach_clicked_cb (GtkWidget*   toolbutton,
     gtk_container_add (GTK_CONTAINER (window), vbox);
     g_object_set_data (G_OBJECT (window), "scrolled", scrolled);
     gtk_window_set_title (GTK_WINDOW (window), title);
-    if (menuitem)
-        gtk_widget_hide (menuitem);
     g_signal_handlers_disconnect_by_func (
         _midori_panel_child_for_scrolled (panel, scrolled),
         midori_panel_widget_destroy_cb, toolitem);
@@ -351,12 +359,23 @@ midori_panel_button_detach_clicked_cb (GtkWidget*   toolbutton,
     toolitem = gtk_toolbar_get_nth_item (GTK_TOOLBAR (panel->toolbar),
                                          n > 0 ? n - 1 : 0);
     if (!gtk_notebook_get_n_pages (GTK_NOTEBOOK (panel->notebook)))
-        gtk_widget_set_sensitive (toolbutton, FALSE);
+        gtk_widget_set_sensitive (GTK_WIDGET (panel->button_detach), FALSE);
     g_signal_connect (window, "delete-event",
         G_CALLBACK (midori_panel_detached_window_delete_event_cb), panel);
     gtk_widget_show (window);
 }
 
+static void
+midori_panel_button_detach_clicked_cb (GtkWidget*   toolbutton,
+                                       MidoriPanel* panel)
+{
+    /* FIXME: What happens when the browser is destroyed? */
+    /* FIXME: What about multiple browsers? */
+    /* FIXME: Should we remember if the child was detached? */
+    gint n = midori_panel_get_current_page (panel);
+    midori_panel_detach_page (panel, n);
+}
+
 static void
 midori_panel_button_align_clicked_cb (GtkWidget*   toolitem,
                                       MidoriPanel* panel)
@@ -386,14 +405,12 @@ midori_panel_init (MidoriPanel* panel)
 
     /* Create the sidebar */
     panel->toolbar = gtk_toolbar_new ();
-    gtk_toolbar_set_style (GTK_TOOLBAR (panel->toolbar), GTK_TOOLBAR_BOTH);
-    gtk_toolbar_set_icon_size (GTK_TOOLBAR (panel->toolbar),
-                               GTK_ICON_SIZE_BUTTON);
-    g_object_set (panel->toolbar, "orientation", GTK_ORIENTATION_VERTICAL, NULL);
-    gtk_box_pack_start (GTK_BOX (panel), panel->toolbar, FALSE, FALSE, 0);
+    gtk_toolbar_set_icon_size (GTK_TOOLBAR (panel->toolbar), GTK_ICON_SIZE_BUTTON);
+    gtk_toolbar_set_show_arrow (GTK_TOOLBAR (panel->toolbar), FALSE);
     gtk_widget_show_all (panel->toolbar);
     vbox = gtk_vbox_new (FALSE, 0);
     gtk_box_pack_start (GTK_BOX (panel), vbox, TRUE, TRUE, 0);
+    gtk_box_pack_end (GTK_BOX (vbox), panel->toolbar, FALSE, FALSE, 0);
 
     /* Create the titlebar */
     labelbar = gtk_toolbar_new ();
@@ -499,17 +516,23 @@ midori_panel_set_property (GObject*      object,
         break;
     case PROP_SHOW_TITLES:
         panel->show_titles = g_value_get_boolean (value);
-        gtk_toolbar_set_style (GTK_TOOLBAR (panel->toolbar),
-            panel->show_titles ? GTK_TOOLBAR_BOTH : GTK_TOOLBAR_ICONS);
+        /* Ignore */
         break;
     case PROP_SHOW_CONTROLS:
         panel->show_controls = g_value_get_boolean (value);
         sokoke_widget_set_visible (panel->labelbar, panel->show_controls);
         sokoke_widget_set_visible (panel->toolbar, panel->show_controls);
+        if (panel->button_controls)
+            gtk_toggle_tool_button_set_active (
+                GTK_TOGGLE_TOOL_BUTTON (panel->button_controls),
+                !panel->show_controls);
         break;
     case PROP_RIGHT_ALIGNED:
         midori_panel_set_right_aligned (panel, g_value_get_boolean (value));
         break;
+    case PROP_OPEN_PANELS_IN_WINDOWS:
+        panel->open_panels_in_windows = g_value_get_boolean (value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;
@@ -548,6 +571,9 @@ midori_panel_get_property (GObject*    object,
     case PROP_RIGHT_ALIGNED:
         g_value_set_boolean (value, panel->right_aligned);
         break;
+    case PROP_OPEN_PANELS_IN_WINDOWS:
+        g_value_set_boolean (value, panel->open_panels_in_windows);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;
@@ -667,65 +693,43 @@ midori_panel_construct_tool_item (MidoriPanel*    panel,
     return GTK_TOOL_ITEM (toolitem);
 }
 
-#if !HAVE_HILDON
-static void
-midori_panel_show_titles_toggled_cb (GtkWidget*   menuitem,
-                                     MidoriPanel* panel)
-{
-    g_object_set (panel, "show-titles", !panel->show_titles, NULL);
-}
-
-static void
-midori_panel_show_controls_toggled_cb (GtkWidget*   menuitem,
-                                       MidoriPanel* panel)
-{
-    g_object_set (panel, "show-controls", !panel->show_controls, NULL);
-}
-
-static void
-midori_panel_options_clicked_cb (GtkToolItem* toolitem,
-                                 MidoriPanel* panel)
-{
-    gint n;
-    GtkWidget* viewable;
-    GtkWidget* menu;
-    GtkWidget* menuitem;
-
-    n = midori_panel_get_current_page (panel);
-    viewable = midori_panel_get_nth_page (panel, n);
-    menu = gtk_menu_new ();
-    menuitem = gtk_check_menu_item_new_with_mnemonic (_("Show panel _titles"));
-    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuitem),
-                                    panel->show_titles);
-    g_signal_connect (menuitem, "toggled",
-        G_CALLBACK (midori_panel_show_titles_toggled_cb), panel);
-    gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
-    gtk_widget_show (menuitem);
-    menuitem = gtk_check_menu_item_new_with_mnemonic (_("Show operating _controls"));
-    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menuitem),
-                                    panel->show_controls);
-    g_signal_connect (menuitem, "toggled",
-        G_CALLBACK (midori_panel_show_controls_toggled_cb), panel);
-    gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
-    gtk_widget_show (menuitem);
-    g_signal_emit_by_name (viewable, "populate-option-menu", menu);
-
-    katze_widget_popup (GTK_WIDGET (toolitem), GTK_MENU (menu),
-                        NULL, SOKOKE_MENU_POSITION_LEFT);
-}
-#endif
-
 static void
 midori_panel_action_activate_cb (GtkRadioAction* action,
                                  MidoriPanel*    panel)
 {
-    gint n = katze_object_get_int (action, "value");
+    GtkWidget* viewable = g_object_get_data (G_OBJECT (action), "viewable");
+    gint n = midori_panel_page_num (panel, viewable);
 
-    midori_panel_set_current_page (panel, n);
-    g_signal_emit (panel, signals[SWITCH_PAGE], 0, n);
-    gtk_widget_show (GTK_WIDGET (panel));
+    /* If the panel is detached, focus the window */
+    if (n == -1)
+    {
+        GtkWidget* toplevel = gtk_widget_get_toplevel (viewable);
+        gtk_window_present (GTK_WINDOW (toplevel));
+        return;
+    }
+
+    if (panel->open_panels_in_windows
+     && gtk_radio_action_get_current_value (action)
+     == katze_object_get_int (action, "value"))
+        midori_panel_detach_page (panel, n);
+    else
+    {
+        midori_panel_set_current_page (panel, n);
+        g_signal_emit (panel, signals[SWITCH_PAGE], 0, n);
+        gtk_widget_show (GTK_WIDGET (panel));
+    }
 }
 
+#if !HAVE_HILDON
+static void
+midori_panel_show_controls_toggled_cb (GtkWidget*   menuitem,
+                                       MidoriPanel* panel)
+{
+    if (panel->show_controls != !gtk_toggle_tool_button_get_active (GTK_TOGGLE_TOOL_BUTTON (menuitem)))
+        g_object_set (panel, "show-controls", !panel->show_controls, NULL);
+}
+#endif
+
 /**
  * midori_panel_append_page:
  * @panel: a #MidoriPanel
@@ -789,12 +793,15 @@ midori_panel_append_page (MidoriPanel*    panel,
 
     toolbar = midori_viewable_get_toolbar (viewable);
     #if !HAVE_HILDON
-    toolitem = gtk_tool_button_new_from_stock (GTK_STOCK_PROPERTIES);
-    gtk_tool_item_set_tooltip_text (toolitem, _("Options"));
+    toolitem = gtk_toggle_tool_button_new_from_stock (GTK_STOCK_PROPERTIES);
+    gtk_tool_item_set_tooltip_text (toolitem, _("Hide operating controls"));
+    gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (toolitem),
+        !panel->show_controls);
     g_signal_connect (toolitem, "clicked",
-        G_CALLBACK (midori_panel_options_clicked_cb), panel);
+        G_CALLBACK (midori_panel_show_controls_toggled_cb), panel);
     gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolitem, 0);
     gtk_widget_show (GTK_WIDGET (toolitem));
+    panel->button_controls = toolitem;
     #endif
     gtk_widget_show (toolbar);
     gtk_container_add (GTK_CONTAINER (panel->toolbook), toolbar);
@@ -807,18 +814,19 @@ midori_panel_append_page (MidoriPanel*    panel,
         midori_viewable_get_stock_id (viewable), NULL);
     action = (GtkAction*)gtk_radio_action_new (action_name,
         midori_viewable_get_label (viewable),
-        NULL, midori_viewable_get_stock_id (viewable), n);
+        midori_viewable_get_label (viewable),
+        midori_viewable_get_stock_id (viewable), n);
+    g_object_set_data (G_OBJECT (action), "viewable", viewable);
     g_signal_connect (action, "activate",
         G_CALLBACK (midori_panel_action_activate_cb), panel);
     if (panel->action_group)
     {
-        /* FIXME: For some reason the accelerator only works if a menuitem
-            is created, but not before that. */
         GtkWidget* toplevel = gtk_widget_get_toplevel (GTK_WIDGET (panel));
         GSList* groups = gtk_accel_groups_from_object (G_OBJECT (toplevel));
         gtk_action_set_accel_group (action, g_slist_nth_data (groups, 0));
         gtk_action_group_add_action_with_accel (panel->action_group,
                                                 action, NULL);
+        gtk_action_connect_accelerator (action);
     }
     if (n > 0)
         g_object_set (action, "group", g_object_get_data (
@@ -979,11 +987,19 @@ midori_panel_set_current_page (MidoriPanel* panel,
 
     if ((viewable = midori_panel_get_nth_page (panel, n)))
     {
+        GtkWidget* toolbar;
+        GList* items;
         const gchar* label;
 
         if (!GTK_WIDGET_VISIBLE (viewable))
             return;
+
         gtk_notebook_set_current_page (GTK_NOTEBOOK (panel->toolbook), n);
+        toolbar = gtk_notebook_get_nth_page (GTK_NOTEBOOK (panel->toolbook), n);
+        items = gtk_container_get_children (GTK_CONTAINER (toolbar));
+        sokoke_widget_set_visible (panel->toolbook,
+            g_list_nth_data (items, 1) != NULL);
+        g_list_free (items);
         gtk_notebook_set_current_page (GTK_NOTEBOOK (panel->notebook), n);
         label = midori_viewable_get_label (MIDORI_VIEWABLE (viewable));
         g_object_set (panel->toolbar_label, "label", label, NULL);
