@@ -50,6 +50,7 @@ struct _MidoriWebSettings
     gboolean compact_sidepanel;
     gboolean show_panel_controls;
     gboolean right_align_sidepanel;
+    gboolean open_panels_in_windows;
 
     MidoriStartup load_on_startup;
     gchar* homepage;
@@ -85,6 +86,7 @@ struct _MidoriWebSettings
     gboolean remember_last_downloaded_files;
 
     gchar* http_proxy;
+    gchar* http_accept_language;
     gboolean auto_detect_proxy;
     MidoriIdentity identify_as;
     gchar* ident_string;
@@ -125,6 +127,7 @@ enum
     PROP_COMPACT_SIDEPANEL,
     PROP_SHOW_PANEL_CONTROLS,
     PROP_RIGHT_ALIGN_SIDEPANEL,
+    PROP_OPEN_PANELS_IN_WINDOWS,
 
     PROP_LOAD_ON_STARTUP,
     PROP_HOMEPAGE,
@@ -151,6 +154,11 @@ enum
     PROP_AUTO_LOAD_IMAGES,
     PROP_ENABLE_SCRIPTS,
     PROP_ENABLE_PLUGINS,
+    PROP_ENABLE_DEVELOPER_EXTRAS,
+    PROP_ENABLE_HTML5_DATABASE,
+    PROP_ENABLE_HTML5_LOCAL_STORAGE,
+    PROP_ENABLE_OFFLINE_WEB_APPLICATION_CACHE,
+    PROP_ENABLE_PAGE_CACHE,
     PROP_ZOOM_TEXT_AND_IMAGES,
     PROP_FIND_WHILE_TYPING,
     PROP_KINETIC_SCROLLING,
@@ -165,7 +173,8 @@ enum
     PROP_HTTP_PROXY,
     PROP_AUTO_DETECT_PROXY,
     PROP_IDENTIFY_AS,
-    PROP_IDENT_STRING,
+    PROP_USER_AGENT,
+    PROP_PREFERRED_LANGUAGES,
 
     PROP_CLEAR_PRIVATE_DATA
 };
@@ -252,6 +261,7 @@ midori_toolbar_style_get_type (void)
         static const GEnumValue values[] = {
          { MIDORI_TOOLBAR_DEFAULT, "MIDORI_TOOLBAR_DEFAULT", N_("Default") },
          { MIDORI_TOOLBAR_ICONS, "MIDORI_TOOLBAR_ICONS", N_("Icons") },
+         { MIDORI_TOOLBAR_SMALL_ICONS, "MIDORI_TOOLBAR_SMALL_ICONS", N_("Small icons") },
          { MIDORI_TOOLBAR_TEXT, "MIDORI_TOOLBAR_TEXT", N_("Text") },
          { MIDORI_TOOLBAR_BOTH, "MIDORI_TOOLBAR_BOTH", N_("Icons and text") },
          { MIDORI_TOOLBAR_BOTH_HORIZ, "MIDORI_TOOLBAR_BOTH_HORIZ", N_("Text beside icons") },
@@ -522,7 +532,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      "toolbar-items",
                                      _("Toolbar Items"),
                                      _("The items to show on the toolbar"),
-                                     "TabNew,Back,Forward,ReloadStop,Location,Panel,Search,Trash",
+                                     "TabNew,Back,Forward,Next,ReloadStop,Location,Panel,Search,Trash",
                                      flags));
 
     g_object_class_install_property (gobject_class,
@@ -566,6 +576,22 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      FALSE,
                                      flags));
 
+    /**
+     * MidoriWebSettings:open-panels-in-window:
+     *
+     * Whether to open panels in separate windows.
+     *
+     * Since: 0.2.2
+     */
+    g_object_class_install_property (gobject_class,
+                                     PROP_OPEN_PANELS_IN_WINDOWS,
+                                     g_param_spec_boolean (
+                                     "open-panels-in-windows",
+                                     _("Open panels in separate windows"),
+        _("Whether to always open panels in separate windows"),
+                                     FALSE,
+                                     flags));
+
 
     g_object_class_install_property (gobject_class,
                                      PROP_LOAD_ON_STARTUP,
@@ -627,7 +653,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("The folder downloaded files are saved to"),
                                      midori_get_download_dir (),
     #if WEBKIT_CHECK_VERSION (1, 1, 3)
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
     #else
                                      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
     #endif
@@ -649,7 +675,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
         _("Whether to ask for the destination folder when downloading a file"),
                                      FALSE,
     #if WEBKIT_CHECK_VERSION (1, 1, 15)
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
     #else
                                      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
     #endif
@@ -669,7 +695,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
         _("Whether to show a notification when a transfer has been completed"),
                                      TRUE,
     #if WEBKIT_CHECK_VERSION (1, 1, 3)
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
     #else
                                      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
     #endif
@@ -705,11 +731,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      "news-aggregator",
                                      _("News Aggregator"),
                                      _("An external news aggregator"),
-                                     #if HAVE_HILDON
-                                     "osso_notes",
-                                     #else
                                      NULL,
-                                     #endif
                                      flags));
 
     g_object_class_install_property (gobject_class,
@@ -718,7 +740,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      "location-entry-search",
                                      _("Location entry Search"),
                                      _("The search to perform inside the location entry"),
-                                     "http://www.google.com/search?q=%s",
+                                     NULL,
                                      flags));
 
     g_object_class_install_property (gobject_class,
@@ -757,12 +779,8 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Open new pages in:"),
                                      _("Where to open new pages"),
                                      MIDORI_TYPE_NEW_PAGE,
-                                     #if HAVE_HILDON
-                                     MIDORI_NEW_PAGE_WINDOW,
-                                     #else
                                      MIDORI_NEW_PAGE_TAB,
-                                     #endif
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
     g_object_class_install_property (gobject_class,
                                      PROP_OPEN_EXTERNAL_PAGES_IN,
@@ -771,12 +789,8 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Open external pages in:"),
                                      _("Where to open externally opened pages"),
                                      MIDORI_TYPE_NEW_PAGE,
-                                     #if HAVE_HILDON
-                                     MIDORI_NEW_PAGE_WINDOW,
-                                     #else
                                      MIDORI_NEW_PAGE_TAB,
-                                     #endif
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
     g_object_class_install_property (gobject_class,
                                      PROP_MIDDLE_CLICK_OPENS_SELECTION,
@@ -812,7 +826,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Open popups in tabs"),
                                      _("Whether to open popup windows in tabs"),
                                      TRUE,
-                                     G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
 
     /* Override properties to localize them for preference proxies */
@@ -840,6 +854,49 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Enable embedded Netscape plugin objects"),
                                      TRUE,
                                      flags));
+    /* Override properties to override defaults */
+    g_object_class_install_property (gobject_class,
+                                     PROP_ENABLE_DEVELOPER_EXTRAS,
+                                     g_param_spec_boolean (
+                                     "enable-developer-extras",
+                                     "Enable developer tools",
+                                     "Enable special extensions for developers",
+                                     TRUE,
+                                     flags));
+    #if WEBKIT_CHECK_VERSION (1, 1, 8)
+    g_object_class_install_property (gobject_class,
+                                     PROP_ENABLE_HTML5_DATABASE,
+                                     g_param_spec_boolean ("enable-html5-database",
+                                                           _("Enable HTML5 database support"),
+                                                           _("Whether to enable HTML5 database support"),
+                                                           FALSE,
+                                                           flags));
+    g_object_class_install_property (gobject_class,
+                                     PROP_ENABLE_HTML5_LOCAL_STORAGE,
+                                     g_param_spec_boolean ("enable-html5-local-storage",
+                                                           _("Enable HTML5 local storage support"),
+                                                           _("Whether to enable HTML5 local storage support"),
+                                                           FALSE,
+                                                           flags));
+    #endif
+    #if WEBKIT_CHECK_VERSION (1, 1, 13)
+    g_object_class_install_property (gobject_class,
+                                     PROP_ENABLE_OFFLINE_WEB_APPLICATION_CACHE,
+                                     g_param_spec_boolean ("enable-offline-web-application-cache",
+                                                           _("Enable offline web application cache"),
+                                                           _("Whether to enable offline web application cache"),
+                                                           FALSE,
+                                                           flags));
+    #endif
+    #if WEBKIT_CHECK_VERSION (1, 1, 18)
+    g_object_class_install_property (gobject_class,
+                                     PROP_ENABLE_PAGE_CACHE,
+                                     g_param_spec_boolean ("enable-page-cache",
+                                                           "Enable page cache",
+                                                           "Whether the page cache should be used",
+                                                           TRUE,
+                                                           flags));
+    #endif
 
     /**
      * MidoriWebSettings:zoom-text-and-images:
@@ -855,7 +912,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                       _("Zoom Text and Images"),
                                       _("Whether to zoom text and images"),
                                       FALSE,
-                                      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                      flags));
 
     /**
     * MidoriWebSettings:find-while-typing:
@@ -897,8 +954,15 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("What type of cookies to accept"),
                                      MIDORI_TYPE_ACCEPT_COOKIES,
                                      MIDORI_ACCEPT_COOKIES_ALL,
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
+    /**
+     * MidoriWebSettings:original-cookies-only:
+     *
+     * Accept cookies from the original website only.
+     *
+     * Deprecated: 0.2.3: This value is not used.
+    */
     g_object_class_install_property (gobject_class,
                                      PROP_ORIGINAL_COOKIES_ONLY,
                                      g_param_spec_boolean (
@@ -906,7 +970,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Original cookies only"),
                                      _("Accept cookies from the original website only"),
                                      FALSE,
-                                     G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
     g_object_class_install_property (gobject_class,
                                      PROP_MAXIMUM_COOKIE_AGE,
@@ -915,10 +979,17 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Maximum cookie age"),
                                      _("The maximum number of days to save cookies for"),
                                      0, G_MAXINT, 30,
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
 
 
+    /**
+    * MidoriWebSettings:remember-last-visited-pages:
+    *
+    * Whether the last visited pages are saved.
+    *
+    * Deprecated: 0.2.2
+    */
     g_object_class_install_property (gobject_class,
                                      PROP_REMEMBER_LAST_VISITED_PAGES,
                                      g_param_spec_boolean (
@@ -944,7 +1015,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Remember last downloaded files"),
                                      _("Whether the last downloaded files are saved"),
                                      TRUE,
-                                     G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
 
 
@@ -955,7 +1026,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Proxy Server"),
                                      _("The proxy server used for HTTP connections"),
                                      NULL,
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
     /**
     * MidoriWebSettings:auto-detect-proxy:
@@ -971,7 +1042,7 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("Detect proxy server automatically"),
         _("Whether to detect the proxy server automatically from the environment"),
                                      TRUE,
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
     /**
     * MidoriWebSettings:identify-as:
@@ -989,23 +1060,39 @@ midori_web_settings_class_init (MidoriWebSettingsClass* class)
                                      _("What to identify as to web pages"),
                                      MIDORI_TYPE_IDENTITY,
                                      MIDORI_IDENT_MIDORI,
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
 
     /**
-    * MidoriWebSettings:ident-string:
-    *
-    * The browser identification string.
-    *
-    * Since: 0.1.2
-    */
+     * MidoriWebSettings:user-agent:
+     *
+     * The browser identification string.
+     *
+     * Since: 0.2.3
+     */
     g_object_class_install_property (gobject_class,
-                                     PROP_IDENT_STRING,
+                                     PROP_USER_AGENT,
                                      g_param_spec_string (
-                                     "ident-string",
+                                     "user-agent",
                                      _("Identification string"),
                                      _("The application identification string"),
                                      NULL,
-                                     G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                                     flags));
+
+    /**
+    * MidoriWebSettings:preferred-languages:
+    *
+    * A comma separated list of languages preferred for rendering multilingual webpages.
+    *
+    * Since: 0.2.3
+    */
+    g_object_class_install_property (gobject_class,
+                                     PROP_PREFERRED_LANGUAGES,
+                                     g_param_spec_string (
+                                     "preferred-languages",
+                                     _("Preferred languages"),
+        _("A comma separated list of languages preferred for rendering multilingual webpages, for example \"de\", \"ru,nl\" or \"en-us;q=1.0, fr-fr;q=0.667\""),
+                                     NULL,
+                                     flags));
 
     /**
      * MidoriWebSettings:clear-private-data:
@@ -1091,7 +1178,7 @@ midori_web_settings_finalize (GObject* object)
     G_OBJECT_CLASS (midori_web_settings_parent_class)->finalize (object);
 }
 
-#if defined (G_OS_UNIX)
+#if defined (G_OS_UNIX) && !HAVE_OSX
 static gchar*
 get_sys_name (void)
 {
@@ -1170,8 +1257,8 @@ generate_ident_string (MidoriIdentity identify_as)
             "Safari/419.3 %s",
                                 os, lang, appname);
     case MIDORI_IDENT_FIREFOX:
-        return g_strdup_printf ("Mozilla/5.0 (%s; U; %s; %s; rv:1.8.1) "
-            "Gecko/20061010 Firefox/2.0 %s",
+        return g_strdup_printf ("Mozilla/5.0 (%s; U; %s; %s; rv:1.9.0.2) "
+            "Gecko/2008092313 Firefox/3.8 %s",
                                 platform, os, lang, appname);
     case MIDORI_IDENT_EXPLORER:
         return g_strdup_printf ("Mozilla/4.0 (compatible; "
@@ -1254,6 +1341,9 @@ midori_web_settings_set_property (GObject*      object,
     case PROP_RIGHT_ALIGN_SIDEPANEL:
         web_settings->right_align_sidepanel = g_value_get_boolean (value);
         break;
+    case PROP_OPEN_PANELS_IN_WINDOWS:
+        web_settings->open_panels_in_windows = g_value_get_boolean (value);
+        break;
 
     case PROP_LOAD_ON_STARTUP:
         web_settings->load_on_startup = g_value_get_enum (value);
@@ -1352,6 +1442,32 @@ midori_web_settings_set_property (GObject*      object,
         g_object_set (web_settings, "WebKitWebSettings::enable-plugins",
                       g_value_get_boolean (value), NULL);
         break;
+    case PROP_ENABLE_DEVELOPER_EXTRAS:
+        g_object_set (web_settings, "WebKitWebSettings::enable-developer-extras",
+                      g_value_get_boolean (value), NULL);
+        break;
+    #if WEBKIT_CHECK_VERSION (1, 1, 8)
+    case PROP_ENABLE_HTML5_DATABASE:
+        g_object_set (web_settings, "WebKitWebSettings::enable-html5-database",
+                      g_value_get_boolean (value), NULL);
+        break;
+    case PROP_ENABLE_HTML5_LOCAL_STORAGE:
+        g_object_set (web_settings, "WebKitWebSettings::enable-html5-local-storage",
+                      g_value_get_boolean (value), NULL);
+        break;
+    #endif
+    #if WEBKIT_CHECK_VERSION (1, 1, 13)
+    case PROP_ENABLE_OFFLINE_WEB_APPLICATION_CACHE:
+        g_object_set (web_settings, "WebKitWebSettings::enable-offline-web-application-cache",
+                      g_value_get_boolean (value), NULL);
+        break;
+    #endif
+    #if WEBKIT_CHECK_VERSION (1, 1, 18)
+    case PROP_ENABLE_PAGE_CACHE:
+        g_object_set (web_settings, "WebKitWebSettings::enable-page-cache",
+                      g_value_get_boolean (value), NULL);
+        break;
+    #endif
     case PROP_ZOOM_TEXT_AND_IMAGES:
         web_settings->zoom_text_and_images = g_value_get_boolean (value);
         break;
@@ -1395,18 +1511,21 @@ midori_web_settings_set_property (GObject*      object,
             katze_assign (web_settings->ident_string, string);
             #if WEBKIT_CHECK_VERSION (1, 1, 11)
             g_object_set (web_settings, "user-agent", string, NULL);
+            #else
+            g_object_notify (object, "user-agent");
             #endif
-            g_object_notify (object, "ident-string");
         }
         break;
-    case PROP_IDENT_STRING:
+    case PROP_USER_AGENT:
         if (web_settings->identify_as == MIDORI_IDENT_CUSTOM)
-        {
             katze_assign (web_settings->ident_string, g_value_dup_string (value));
-            #if WEBKIT_CHECK_VERSION (1, 1, 11)
-            g_object_set (web_settings, "user-agent", web_settings->ident_string, NULL);
-            #endif
-        }
+        #if WEBKIT_CHECK_VERSION (1, 1, 11)
+        g_object_set (web_settings, "WebKitWebSettings::user-agent",
+                                    web_settings->ident_string, NULL);
+        #endif
+        break;
+    case PROP_PREFERRED_LANGUAGES:
+        katze_assign (web_settings->http_accept_language, g_value_dup_string (value));
         break;
     case PROP_CLEAR_PRIVATE_DATA:
         web_settings->clear_private_data = g_value_get_int (value);
@@ -1489,6 +1608,9 @@ midori_web_settings_get_property (GObject*    object,
     case PROP_RIGHT_ALIGN_SIDEPANEL:
         g_value_set_boolean (value, web_settings->right_align_sidepanel);
         break;
+    case PROP_OPEN_PANELS_IN_WINDOWS:
+        g_value_set_boolean (value, web_settings->open_panels_in_windows);
+        break;
 
     case PROP_LOAD_ON_STARTUP:
         g_value_set_enum (value, web_settings->load_on_startup);
@@ -1564,6 +1686,32 @@ midori_web_settings_get_property (GObject*    object,
         g_value_set_boolean (value, katze_object_get_boolean (web_settings,
                              "WebKitWebSettings::enable-plugins"));
         break;
+    case PROP_ENABLE_DEVELOPER_EXTRAS:
+        g_value_set_boolean (value, katze_object_get_boolean (web_settings,
+                             "WebKitWebSettings::enable-developer-extras"));
+        break;
+    #if WEBKIT_CHECK_VERSION (1, 1, 8)
+    case PROP_ENABLE_HTML5_DATABASE:
+        g_value_set_boolean (value, katze_object_get_boolean (web_settings,
+                             "WebKitWebSettings::enable-html5-database"));
+        break;
+    case PROP_ENABLE_HTML5_LOCAL_STORAGE:
+        g_value_set_boolean (value, katze_object_get_boolean (web_settings,
+                             "WebKitWebSettings::enable-html5-local-storage"));
+        break;
+    #endif
+    #if WEBKIT_CHECK_VERSION (1, 1, 13)
+    case PROP_ENABLE_OFFLINE_WEB_APPLICATION_CACHE:
+        g_value_set_boolean (value, katze_object_get_boolean (web_settings,
+                             "WebKitWebSettings::enable-offline-web-application-cache"));
+        break;
+    #endif
+    #if WEBKIT_CHECK_VERSION (1, 1, 18)
+    case PROP_ENABLE_PAGE_CACHE:
+        g_value_set_boolean (value, katze_object_get_boolean (web_settings,
+                             "WebKitWebSettings::enable-page-cache"));
+        break;
+    #endif
     case PROP_ZOOM_TEXT_AND_IMAGES:
         g_value_set_boolean (value, web_settings->zoom_text_and_images);
         break;
@@ -1602,7 +1750,7 @@ midori_web_settings_get_property (GObject*    object,
     case PROP_IDENTIFY_AS:
         g_value_set_enum (value, web_settings->identify_as);
         break;
-    case PROP_IDENT_STRING:
+    case PROP_USER_AGENT:
         if (!g_strcmp0 (web_settings->ident_string, ""))
         {
             gchar* string = generate_ident_string (web_settings->identify_as);
@@ -1610,6 +1758,9 @@ midori_web_settings_get_property (GObject*    object,
         }
         g_value_set_string (value, web_settings->ident_string);
         break;
+    case PROP_PREFERRED_LANGUAGES:
+        g_value_set_string (value, web_settings->http_accept_language);
+        break;
     case PROP_CLEAR_PRIVATE_DATA:
         g_value_set_int (value, web_settings->clear_private_data);
         break;
