@@ -43,6 +43,7 @@ enum
     MIDORI_CLEAR_WEBSITE_ICONS = 8,
     MIDORI_CLEAR_TRASH = 16,
     MIDORI_CLEAR_ON_QUIT = 32,
+    MIDORI_CLEAR_WEB_CACHE = 64,
 };
 
 typedef enum
@@ -106,6 +107,7 @@ typedef enum
 {
     MIDORI_TOOLBAR_DEFAULT,
     MIDORI_TOOLBAR_ICONS,
+    MIDORI_TOOLBAR_SMALL_ICONS,
     MIDORI_TOOLBAR_TEXT,
     MIDORI_TOOLBAR_BOTH,
     MIDORI_TOOLBAR_BOTH_HORIZ
@@ -147,7 +149,7 @@ midori_identity_get_type (void) G_GNUC_CONST;
     (midori_identity_get_type ())
 
 GType
-midori_web_settings_get_type               (void);
+midori_web_settings_get_type               (void) G_GNUC_CONST;
 
 MidoriWebSettings*
 midori_web_settings_new                    (void);
