@@ -13,12 +13,49 @@
 #ifndef __SOKOKE_H__
 #define __SOKOKE_H__ 1
 
+/* Common behavior modifiers */
+#define MIDORI_MOD_NEW_WINDOW(state) (state & GDK_SHIFT_MASK)
+#define MIDORI_MOD_NEW_TAB(state) (state & GDK_CONTROL_MASK)
+#define MIDORI_MOD_BACKGROUND(state) (state & GDK_SHIFT_MASK)
+#define MIDORI_MOD_SCROLL(state) (state & GDK_CONTROL_MASK)
+
 #include <katze/katze.h>
 
-#include <gtk/gtk.h>
-#include <libsoup/soup.h>
+#include <webkit/webkit.h>
 #include <JavaScriptCore/JavaScript.h>
 
+#if !GLIB_CHECK_VERSION (2, 14, 0)
+    #define G_PARAM_STATIC_STRINGS \
+    (G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB)
+    #define gtk_dialog_get_content_area(dlg) dlg->vbox
+#endif
+
+#if !GTK_CHECK_VERSION (2, 16, 0)
+    #define GTK_ACTIVATABLE GTK_WIDGET
+    #define gtk_activatable_get_related_action gtk_widget_get_action
+#endif
+
+#if !GTK_CHECK_VERSION (2, 18, 0)
+    #define gtk_widget_is_toplevel(widget) GTK_WIDGET_TOPLEVEL (widget)
+    #define gtk_widget_has_focus(widget) GTK_WIDGET_HAS_FOCUS (widget)
+#endif
+
+#if !GTK_CHECK_VERSION(2, 12, 0)
+
+void
+gtk_widget_set_has_tooltip             (GtkWidget*         widget,
+                                        gboolean           has_tooltip);
+
+void
+gtk_widget_set_tooltip_text            (GtkWidget*         widget,
+                                        const gchar*       text);
+
+void
+gtk_tool_item_set_tooltip_text         (GtkToolItem*       toolitem,
+                                        const gchar*       text);
+
+#endif
+
 gchar*
 sokoke_js_script_eval                   (JSContextRef    js_context,
                                          const gchar*    script,
@@ -27,6 +64,18 @@ sokoke_js_script_eval                   (JSContextRef    js_context,
 /* Many themes need this hack for small toolbars to work */
 #define GTK_ICON_SIZE_SMALL_TOOLBAR GTK_ICON_SIZE_BUTTON
 
+void
+sokoke_message_dialog                   (GtkMessageType  message_type,
+                                         const gchar*    short_message,
+                                         const gchar*    detailed_message);
+
+gboolean
+sokoke_show_uri_with_mime_type          (GdkScreen*      screen,
+                                         const gchar*    uri,
+                                         const gchar*    mime_type,
+                                         guint32         timestamp,
+                                         GError**        error);
+
 gboolean
 sokoke_show_uri                         (GdkScreen*      screen,
                                          const gchar*    uri,
@@ -49,18 +98,11 @@ gchar*
 sokoke_uri_to_ascii                     (const gchar*    uri);
 
 gchar*
-sokoke_magic_uri                        (const gchar*    uri,
-                                         KatzeArray*     search_engines);
+sokoke_magic_uri                        (const gchar*    uri);
 
 gchar*
 sokoke_format_uri_for_display           (const gchar*    uri);
 
-typedef enum {
-    SOKOKE_MENU_POSITION_CURSOR = 0,
-    SOKOKE_MENU_POSITION_LEFT,
-    SOKOKE_MENU_POSITION_RIGHT
-} SokokeMenuPos;
-
 void
 sokoke_combo_box_add_strings            (GtkComboBox*    combobox,
                                          const gchar*    label_first,
@@ -73,12 +115,6 @@ sokoke_widget_set_visible               (GtkWidget*      widget,
 void
 sokoke_container_show_children          (GtkContainer*   container);
 
-void
-sokoke_widget_popup                     (GtkWidget*      widget,
-                                         GtkMenu*        menu,
-                                         GdkEventButton* event,
-                                         SokokeMenuPos   pos);
-
 GtkWidget*
 sokoke_xfce_header_new                  (const gchar*    icon,
                                          const gchar*    title);
@@ -142,12 +178,13 @@ sokoke_widget_get_text_size             (GtkWidget*      widget,
 GtkWidget*
 sokoke_action_create_popup_menu_item    (GtkAction*      action);
 
-GtkWidget*
-sokoke_image_menu_item_new_ellipsized   (const gchar*    label);
-
 gint64
 sokoke_time_t_to_julian                 (const time_t*   timestamp);
 
+gint
+sokoke_days_between                     (const time_t*   day1,
+                                         const time_t*   day2);
+
 void
 sokoke_register_stock_items             (void);
 
@@ -165,8 +202,13 @@ sokoke_find_config_filename             (const gchar*    folder,
 gchar*
 sokoke_find_data_filename               (const gchar*    filename);
 
+gchar**
+sokoke_get_argv                         (gchar**         argument_vector);
+
+#if !WEBKIT_CHECK_VERSION (1, 1, 14)
 SoupServer*
 sokoke_get_res_server                   (void);
+#endif
 
 gchar*
 sokoke_replace_variables                (const gchar* template,
@@ -176,4 +218,15 @@ gboolean
 sokoke_window_activate_key              (GtkWindow*      window,
                                          GdkEventKey*    event);
 
+GtkWidget*
+sokoke_file_chooser_dialog_new          (const gchar*         title,
+                                         GtkWindow*           window,
+                                         GtkFileChooserAction action);
+
+gboolean
+sokoke_prefetch_uri                     (const char* uri);
+
+gchar *
+sokoke_accept_languages                 (const gchar* const * lang_names);
+
 #endif /* !__SOKOKE_H__ */
