@@ -18,6 +18,7 @@
 #include "midori-viewable.h"
 
 #include "sokoke.h"
+#include "gtkiconentry.h"
 
 #include <glib/gi18n.h>
 #include <string.h>
@@ -31,6 +32,14 @@ midori_browser_edit_bookmark_dialog_new (MidoriBrowser* browser,
                                          gboolean       new_bookmark,
                                          gboolean       is_folder);
 
+#include "config.h"
+
+#if HAVE_SQLITE
+    #include <sqlite3.h>
+#endif
+
+#define COMPLETION_DELAY 150
+
 struct _MidoriHistory
 {
     GtkVBox parent_instance;
@@ -42,7 +51,9 @@ struct _MidoriHistory
     GtkWidget* treeview;
     MidoriApp* app;
     KatzeArray* array;
-    KatzeNet* net;
+
+    gint filter_timeout;
+    gchar* filter;
 };
 
 struct _MidoriHistoryClass
@@ -114,6 +125,214 @@ midori_history_get_stock_id (MidoriViewable* viewable)
     return STOCK_HISTORY;
 }
 
+#if HAVE_SQLITE
+static void
+midori_history_clear_db (MidoriHistory* history)
+{
+    gchar* sqlcmd;
+    sqlite3* db;
+    char* errmsg = NULL;
+
+    db = g_object_get_data (G_OBJECT (history->array), "db");
+    sqlcmd = sqlite3_mprintf ("DELETE FROM history");
+
+    if (sqlite3_exec (db, sqlcmd, NULL, NULL, &errmsg) != SQLITE_OK)
+    {
+        g_printerr (_("Failed to remove history item: %s\n"), errmsg);
+        sqlite3_free (errmsg);
+    }
+
+    sqlite3_free (sqlcmd);
+}
+
+static void
+midori_history_remove_item_from_db (MidoriHistory* history,
+                                    KatzeItem*     item)
+{
+    gchar* sqlcmd;
+    sqlite3* db;
+    char* errmsg = NULL;
+
+    db = g_object_get_data (G_OBJECT (history->array), "db");
+
+    if (katze_item_get_uri (item))
+        sqlcmd = sqlite3_mprintf (
+            "DELETE FROM history WHERE uri = '%q' AND"
+            " title = '%q' AND date = %llu",
+            katze_item_get_uri (item),
+            katze_item_get_name (item),
+            katze_item_get_added (item));
+    else
+       sqlcmd = sqlite3_mprintf (
+            "DELETE FROM history WHERE day = %d", katze_item_get_added (item));
+
+    if (sqlite3_exec (db, sqlcmd, NULL, NULL, &errmsg) != SQLITE_OK)
+    {
+        g_printerr (_("Failed to remove history item: %s\n"), errmsg);
+        sqlite3_free (errmsg);
+    }
+
+    sqlite3_free (sqlcmd);
+}
+
+/**
+ * midori_history_read_from_db:
+ * @history: a #MidoriHistory
+ * @model: a #GtkTreeStore
+ * @parent: a #GtkTreeIter, or %NULL
+ * @req_day: the timestamp of one day, or 0
+ * @filter: a filter string to search for
+ *
+ * Populates the model according to parameters:
+ * 1. If @req_day is 0, all dates are added as folders.
+ * 2. If @req_day is given, all pages for that day are added.
+ * 3. If @filter is given, all pages matching the filter are added.
+ **/
+static gboolean
+midori_history_read_from_db (MidoriHistory* history,
+                             GtkTreeStore*  model,
+                             GtkTreeIter*   parent,
+                             int            req_day,
+                             const gchar*   filter)
+{
+    sqlite3* db;
+    sqlite3_stmt* statement;
+    gint result;
+    const gchar* sqlcmd;
+    time_t current_time;
+
+    GtkTreeIter iter;
+    GtkTreeIter root_iter;
+
+    db = g_object_get_data (G_OBJECT (history->array), "db");
+
+    if (filter && *filter)
+    {
+        gchar* filterstr;
+
+        sqlcmd = "SELECT uri, title, date "
+                 "FROM history WHERE uri LIKE ? or title LIKE ? "
+                 "GROUP BY uri ORDER BY date ASC";
+        result = sqlite3_prepare_v2 (db, sqlcmd, -1, &statement, NULL);
+        filterstr = g_strdup_printf ("%%%s%%", filter);
+        sqlite3_bind_text (statement, 1, filterstr, -1, g_free);
+        sqlite3_bind_text (statement, 2, g_strdup (filterstr), -1, g_free);
+        req_day = -1;
+    }
+    else if (req_day == 0)
+    {
+        sqlcmd = "SELECT day, date FROM history GROUP BY day ORDER BY day ASC";
+        result = sqlite3_prepare_v2 (db, sqlcmd, -1, &statement, NULL);
+    }
+    else
+    {
+        sqlcmd = "SELECT uri, title, date, day "
+                 "FROM history WHERE day = ? "
+                 "GROUP BY uri ORDER BY date ASC";
+        result = sqlite3_prepare_v2 (db, sqlcmd, -1, &statement, NULL);
+        sqlite3_bind_int64 (statement, 1, req_day);
+    }
+
+    if (result != SQLITE_OK)
+        return FALSE;
+
+    if (req_day == 0)
+        current_time = time (NULL);
+
+    while ((result = sqlite3_step (statement)) == SQLITE_ROW)
+    {
+        KatzeItem* item;
+        const unsigned char* uri;
+        const unsigned char* title;
+        sqlite3_int64 date;
+        sqlite3_int64 day;
+
+        if (req_day == 0)
+        {
+            gint age;
+            gchar token[50];
+            gchar* sdate;
+
+            day = sqlite3_column_int64 (statement, 0);
+            date = sqlite3_column_int64 (statement, 1);
+
+            item = katze_item_new ();
+            katze_item_set_added (item, day);
+            age = sokoke_days_between ((time_t*)&date, &current_time);
+
+            /* A negative age is a date in the future, the clock is probably off */
+            if (age < -1)
+            {
+                static gboolean clock_warning = FALSE;
+                if (!clock_warning)
+                {
+                    midori_app_send_notification (history->app,
+                        _("Erroneous clock time"),
+                        _("The clock time lies in the past. "
+                          "Please check the current date and time."));
+                    clock_warning = TRUE;
+                }
+            }
+
+            if (age > 7 || age < 0)
+            {
+                strftime (token, sizeof (token), "%x", localtime ((time_t*)&date));
+                sdate = token;
+            }
+            else if (age > 6)
+                sdate = _("A week ago");
+            else if (age > 1)
+                sdate = g_strdup_printf (ngettext ("%d day ago",
+                    "%d days ago", (gint)age), (gint)age);
+            else if (age == 0)
+                sdate = _("Today");
+            else
+                sdate = _("Yesterday");
+
+            gtk_tree_store_insert_with_values (model, &root_iter, NULL,
+                                               0, 0, item, 1, sdate, -1);
+            /* That's an invisible dummy, so we always have an expander */
+            gtk_tree_store_insert_with_values (model, &iter, &root_iter,
+                0, 0, NULL, 1, NULL, -1);
+
+            if (age > 1 && age < 7)
+                g_free (sdate);
+        }
+        else
+        {
+            uri = sqlite3_column_text (statement, 0);
+            title = sqlite3_column_text (statement, 1);
+            date = sqlite3_column_int64 (statement, 2);
+            day = sqlite3_column_int64 (statement, 3);
+            if (!uri)
+                continue;
+
+            item = katze_item_new ();
+            katze_item_set_added (item, date);
+            katze_item_set_uri (item, (gchar*)uri);
+            katze_item_set_name (item, (gchar*)title);
+            gtk_tree_store_insert_with_values (model, NULL, parent,
+                0, 0, item, 1, katze_item_get_name (item), -1);
+        }
+    }
+
+    if (req_day != 0 && !(filter && *filter))
+    {
+        /* Remove invisible dummy row */
+        GtkTreeIter child;
+        gint last = gtk_tree_model_iter_n_children (GTK_TREE_MODEL (model), parent);
+        gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (model), &child, parent, last - 1);
+        gtk_tree_store_remove (model, &child);
+    }
+
+    if (result != SQLITE_DONE)
+        g_print (_("Failed to execute database statement: %s\n"),
+                 sqlite3_errmsg (db));
+
+    sqlite3_finalize (statement);
+    return FALSE;
+}
+
 static void
 midori_history_add_clicked_cb (GtkWidget* toolitem)
 {
@@ -133,16 +352,10 @@ midori_history_delete_clicked_cb (GtkWidget*     toolitem,
                                            &model, &iter))
     {
         KatzeItem* item;
-        KatzeArray* parent;
 
         gtk_tree_model_get (model, &iter, 0, &item, -1);
-
-        /* FIXME: Even toplevel items should technically have a parent */
-        g_return_if_fail (katze_item_get_parent (item));
-
-        parent = katze_item_get_parent (item);
-        katze_array_remove_item (parent, item);
-
+        midori_history_remove_item_from_db (history, item);
+        gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
         g_object_unref (item);
     }
 }
@@ -165,7 +378,7 @@ midori_history_clear_clicked_cb (GtkWidget*     toolitem,
     if (result != GTK_RESPONSE_YES)
         return;
 
-    katze_array_clear (history->array);
+    midori_history_clear_db (history);
 }
 
 static void
@@ -185,11 +398,12 @@ midori_history_cursor_or_row_changed_cb (GtkTreeView*   treeview,
 
         gtk_tree_model_get (model, &iter, 0, &item, -1);
 
-        is_page = !KATZE_IS_ARRAY (item) && katze_item_get_uri (item);
+        is_page = item && katze_item_get_uri (item);
         gtk_widget_set_sensitive (history->bookmark, is_page);
         gtk_widget_set_sensitive (history->delete, TRUE);
 
-        g_object_unref (item);
+        if (item)
+            g_object_unref (item);
     }
     else
     {
@@ -197,6 +411,7 @@ midori_history_cursor_or_row_changed_cb (GtkTreeView*   treeview,
         gtk_widget_set_sensitive (history->delete, FALSE);
     }
 }
+#endif
 
 static GtkWidget*
 midori_history_get_toolbar (MidoriViewable* viewable)
@@ -206,12 +421,15 @@ midori_history_get_toolbar (MidoriViewable* viewable)
     if (!history->toolbar)
     {
         GtkWidget* toolbar;
+        #if HAVE_SQLITE
         GtkToolItem* toolitem;
+        #endif
 
         toolbar = gtk_toolbar_new ();
         gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_BOTH_HORIZ);
         gtk_toolbar_set_icon_size (GTK_TOOLBAR (toolbar), GTK_ICON_SIZE_BUTTON);
         history->toolbar = toolbar;
+        #if HAVE_SQLITE
         toolitem = gtk_tool_button_new_from_stock (STOCK_BOOKMARK_ADD);
         gtk_widget_set_tooltip_text (GTK_WIDGET (toolitem),
                                      _("Bookmark the selected history item"));
@@ -245,6 +463,7 @@ midori_history_get_toolbar (MidoriViewable* viewable)
             G_CALLBACK (gtk_widget_destroyed), &history->delete);
         g_signal_connect (history->clear, "destroy",
             G_CALLBACK (gtk_widget_destroyed), &history->clear);
+        #endif
     }
 
     return history->toolbar;
@@ -258,222 +477,6 @@ midori_history_viewable_iface_init (MidoriViewableIface* iface)
     iface->get_toolbar = midori_history_get_toolbar;
 }
 
-static void
-midori_history_add_item_cb (KatzeArray*    array,
-                            KatzeItem*     added_item,
-                            MidoriHistory* history);
-
-static void
-midori_history_remove_item_cb (KatzeArray*    array,
-                               KatzeItem*     removed_item,
-                               MidoriHistory* history);
-
-static void
-midori_history_clear_cb (KatzeArray*    array,
-                         MidoriHistory* history);
-
-static void
-midori_history_disconnect_folder (MidoriHistory* history,
-                                  KatzeArray*    array,
-                                  gboolean       unref)
-{
-    KatzeItem* item;
-    guint i;
-
-    g_return_if_fail (KATZE_IS_ARRAY (array));
-
-    g_signal_handlers_disconnect_by_func (array,
-        midori_history_add_item_cb, history);
-    g_signal_handlers_disconnect_by_func (array,
-        midori_history_remove_item_cb, history);
-    g_signal_handlers_disconnect_by_func (array,
-        midori_history_clear_cb, history);
-
-    i = 0;
-    while ((item = katze_array_get_nth_item (array, i++)))
-    {
-        if (KATZE_IS_ARRAY (item))
-            midori_history_disconnect_folder (history, KATZE_ARRAY (item), unref);
-        if (unref)
-            g_object_unref (item);
-    }
-}
-
-static void
-midori_history_add_item_cb (KatzeArray*    array,
-                            KatzeItem*     added_item,
-                            MidoriHistory* history)
-{
-    GtkTreeModel* model;
-    GtkTreeIter iter;
-    guint i;
-
-    g_return_if_fail (KATZE_IS_ARRAY (array));
-    g_return_if_fail (KATZE_IS_ITEM (added_item));
-    g_return_if_fail (MIDORI_IS_HISTORY (history));
-
-    if (KATZE_IS_ARRAY (added_item))
-    {
-        g_signal_connect (added_item, "add-item",
-            G_CALLBACK (midori_history_add_item_cb), history);
-        g_signal_connect (added_item, "remove-item",
-            G_CALLBACK (midori_history_remove_item_cb), history);
-        g_signal_connect (added_item, "clear",
-            G_CALLBACK (midori_history_clear_cb), history);
-    }
-
-    g_object_ref (added_item);
-    model = gtk_tree_view_get_model (GTK_TREE_VIEW (history->treeview));
-
-    if (array == history->array)
-    {
-        gtk_tree_store_insert_with_values (GTK_TREE_STORE (model),
-            &iter, NULL, 0, 0, added_item, -1);
-        return;
-    }
-
-    i = 0;
-    /* FIXME: Recurse over children of folders, too */
-    while (gtk_tree_model_iter_nth_child (model, &iter, NULL, i))
-    {
-        KatzeItem* item;
-
-        gtk_tree_model_get (model, &iter, 0, &item, -1);
-        if (item == (KatzeItem*)array)
-        {
-            GtkTreeIter child_iter;
-
-            gtk_tree_store_insert_with_values (GTK_TREE_STORE (model),
-                &child_iter, &iter, 0, 0, added_item, -1);
-            break;
-        }
-        g_object_unref (item);
-
-        i++;
-    }
-}
-
-static void
-midori_history_remove_iter (GtkTreeModel* model,
-                            GtkTreeIter*  parent,
-                            KatzeItem*    removed_item)
-{
-    guint i;
-    GtkTreeIter iter;
-
-    i = 0;
-    while (gtk_tree_model_iter_nth_child (model, &iter, parent, i))
-    {
-        KatzeItem* item;
-
-        gtk_tree_model_get (model, &iter, 0, &item, -1);
-
-        if (item == removed_item)
-        {
-            gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
-            g_object_unref (item);
-            break;
-        }
-
-        if (KATZE_IS_ARRAY (item))
-            midori_history_remove_iter (model, &iter, removed_item);
-
-        g_object_unref (item);
-        i++;
-    }
-}
-
-static void
-midori_history_remove_item_cb (KatzeArray*    array,
-                               KatzeItem*     removed_item,
-                               MidoriHistory* history)
-{
-    GtkTreeModel* model;
-
-    g_assert (KATZE_IS_ARRAY (array));
-    g_assert (KATZE_IS_ITEM (removed_item));
-
-    if (KATZE_IS_ARRAY (removed_item))
-        midori_history_disconnect_folder (history, KATZE_ARRAY (removed_item), TRUE);
-
-    model = gtk_tree_view_get_model (GTK_TREE_VIEW (history->treeview));
-    midori_history_remove_iter (model, NULL, removed_item);
-    g_object_unref (removed_item);
-}
-
-static void
-midori_history_clear_cb (KatzeArray*    array,
-                         MidoriHistory* history)
-{
-    GtkTreeView* treeview;
-    GtkTreeStore* store;
-
-    g_assert (KATZE_IS_ARRAY (array));
-
-    if (array == history->array)
-    {
-        treeview = GTK_TREE_VIEW (history->treeview);
-        store = GTK_TREE_STORE (gtk_tree_view_get_model (treeview));
-        gtk_tree_store_clear (store);
-    }
-    else
-    {
-        KatzeItem* item;
-        guint i;
-
-        i = 0;
-        while ((item = katze_array_get_nth_item (array, i++)))
-            midori_history_remove_item_cb (array, item, history);
-    }
-
-    midori_history_disconnect_folder (history, array, TRUE);
-}
-
-static void
-midori_history_insert_item (MidoriHistory* history,
-                            GtkTreeStore*  treestore,
-                            GtkTreeIter*   parent,
-                            KatzeItem*     item,
-                            gint64         day)
-{
-    GtkTreeIter iter;
-    gint64 age = -1;
-
-    g_return_if_fail (KATZE_IS_ITEM (item));
-
-    if (KATZE_IS_ARRAY (item))
-    {
-        GtkTreeIter* piter;
-        gint64 pday;
-        guint i;
-        KatzeItem* child;
-
-        g_signal_connect (item, "add-item",
-            G_CALLBACK (midori_history_add_item_cb), history);
-        g_signal_connect (item, "remove-item",
-            G_CALLBACK (midori_history_remove_item_cb), history);
-        g_signal_connect (item, "clear",
-            G_CALLBACK (midori_history_clear_cb), history);
-
-        piter = parent;
-        if ((pday = katze_item_get_added (item)))
-        {
-            age = day - pday;
-            gtk_tree_store_insert_with_values (treestore, &iter, parent,
-                                               0, 0, item, 1, age, -1);
-            piter = &iter;
-        }
-        i = 0;
-        while ((child = katze_array_get_nth_item (KATZE_ARRAY (item), i++)))
-            midori_history_insert_item (history, treestore, piter, child, day);
-    }
-    else
-    {
-        gtk_tree_store_insert_with_values (treestore, &iter, parent,
-                                           0, 0, item, 1, age, -1);
-    }
-}
-
 static void
 midori_history_set_app (MidoriHistory* history,
                         MidoriApp*     app)
@@ -482,26 +485,22 @@ midori_history_set_app (MidoriHistory* history,
 
     if (history->array)
     {
-        midori_history_disconnect_folder (history, history->array, TRUE);
         g_object_unref (history->array);
         model = gtk_tree_view_get_model (GTK_TREE_VIEW (history->treeview));
         gtk_tree_store_clear (GTK_TREE_STORE (model));
     }
+
     katze_assign (history->app, app);
     if (!app)
         return;
-
     g_object_ref (app);
-    history->array = katze_object_get_object (app, "history");
-    if (history->array)
-    {
-        time_t now = time (NULL);
-        gint64 day = sokoke_time_t_to_julian (&now);
 
-        model = gtk_tree_view_get_model (GTK_TREE_VIEW (history->treeview));
-        midori_history_insert_item (history, GTK_TREE_STORE (model),
-            NULL, KATZE_ITEM (history->array), day);
-    }
+    history->array = katze_object_get_object (app, "history");
+    model = gtk_tree_view_get_model (GTK_TREE_VIEW (history->treeview));
+    #if HAVE_SQLITE
+    if (history->array)
+        midori_history_read_from_db (history, GTK_TREE_STORE (model), NULL, 0, NULL);
+    #endif
 }
 
 static void
@@ -550,89 +549,28 @@ midori_history_treeview_render_icon_cb (GtkTreeViewColumn* column,
                                         GtkWidget*         treeview)
 {
     KatzeItem* item;
-    GdkPixbuf* pixbuf = NULL;
+    GdkPixbuf* pixbuf;
 
     gtk_tree_model_get (model, iter, 0, &item, -1);
 
-    g_assert (KATZE_IS_ITEM (item));
-
-    if (KATZE_IS_ARRAY (item))
+    if (!item)
+        pixbuf = NULL;
+    else if (katze_item_get_uri (item))
+        pixbuf = katze_load_cached_icon (katze_item_get_uri (item), treeview);
+    else
         pixbuf = gtk_widget_render_icon (treeview, GTK_STOCK_DIRECTORY,
                                          GTK_ICON_SIZE_MENU, NULL);
-    else
-        pixbuf = katze_load_cached_icon (katze_item_get_uri (item), treeview);
 
     g_object_set (renderer, "pixbuf", pixbuf, NULL);
 
     if (pixbuf)
-        g_object_unref (pixbuf);
-
-    g_object_unref (item);
-}
-
-static void
-midori_history_treeview_render_text_cb (GtkTreeViewColumn* column,
-                                        GtkCellRenderer*   renderer,
-                                        GtkTreeModel*      model,
-                                        GtkTreeIter*       iter,
-                                        MidoriHistory*     history)
-{
-    KatzeItem* item;
-    gint64 age;
-
-    gtk_tree_model_get (model, iter, 0, &item, 1, &age, -1);
-
-    g_assert (KATZE_IS_ITEM (item));
-
-    if (KATZE_IS_ARRAY (item))
     {
-        gchar* sdate;
-
-        /* A negative age is a date in the future, the clock is probably off */
-        if (age < -1)
-        {
-            static gboolean clock_warning = FALSE;
-            if (!clock_warning)
-            {
-                midori_app_send_notification (history->app,
-                    _("Erroneous clock time"),
-                    _("The clock time lies in the past. "
-                      "Please check the current date and time."));
-                clock_warning = TRUE;
-            }
-        }
-
-        if (age > 7 || age < 0)
-        {
-            g_object_set (renderer, "text", katze_item_get_name (item), NULL);
-        }
-        else if (age > 6)
-        {
-            sdate = _("A week ago");
-            g_object_set (renderer, "text", sdate, NULL);
-        }
-        else if (age > 1)
-        {
-            sdate = g_strdup_printf (ngettext ("%d day ago",
-                "%d days ago", (gint)age), (gint)age);
-            g_object_set (renderer, "text", sdate, NULL);
-            g_free (sdate);
-        }
-        else
-        {
-            if (age == 0)
-                sdate = _("Today");
-            else
-                sdate = _("Yesterday");
-            g_object_set (renderer, "text", sdate, NULL);
-        }
+        g_object_unref (pixbuf);
+        g_object_unref (item);
     }
-    else
-        g_object_set (renderer, "text", katze_item_get_name (item), NULL);
-
-    g_object_unref (item);
 }
 
+#if HAVE_SQLITE
 static void
 midori_history_row_activated_cb (GtkTreeView*       treeview,
                                    GtkTreePath*       path,
@@ -649,6 +587,10 @@ midori_history_row_activated_cb (GtkTreeView*       treeview,
     if (gtk_tree_model_get_iter (model, &iter, path))
     {
         gtk_tree_model_get (model, &iter, 0, &item, -1);
+
+        if (!item)
+            return;
+
         uri = katze_item_get_uri (item);
         if (uri && *uri)
         {
@@ -791,22 +733,6 @@ midori_history_bookmark_activate_cb (GtkWidget*     menuitem,
     }
 }
 
-static void
-midori_history_delete_activate_cb (GtkWidget*     menuitem,
-                                   MidoriHistory* history)
-{
-    KatzeItem* item;
-    KatzeArray* parent;
-
-    item = (KatzeItem*)g_object_get_data (G_OBJECT (menuitem), "KatzeItem");
-
-    /* FIXME: Even toplevel items should technically have a parent */
-    g_return_if_fail (katze_item_get_parent (item));
-
-    parent = katze_item_get_parent (item);
-    katze_array_remove_item (parent, item);
-}
-
 static void
 midori_history_popup (GtkWidget*      widget,
                       GdkEventButton* event,
@@ -836,7 +762,7 @@ midori_history_popup (GtkWidget*      widget,
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
     gtk_widget_show (menuitem);
     midori_history_popup_item (menu, GTK_STOCK_DELETE, NULL,
-        item, midori_history_delete_activate_cb, history);
+        item, midori_history_delete_clicked_cb, history);
 
     katze_widget_popup (widget, GTK_MENU (menu), event, KATZE_MENU_POSITION_CURSOR);
 }
@@ -858,6 +784,9 @@ midori_history_button_release_event_cb (GtkWidget*      widget,
 
         gtk_tree_model_get (model, &iter, 0, &item, -1);
 
+        if (!item)
+            return FALSE;
+
         if (event->button == 2)
         {
             const gchar* uri = katze_item_get_uri (item);
@@ -895,13 +824,10 @@ midori_history_key_release_event_cb (GtkWidget*     widget,
     if (katze_tree_view_get_selected_iter (GTK_TREE_VIEW (widget), &model, &iter))
     {
         KatzeItem* item;
-        KatzeArray* parent;
 
         gtk_tree_model_get (model, &iter, 0, &item, -1);
-
-        parent = katze_item_get_parent (item);
-        katze_array_remove_item (parent, item);
-
+        midori_history_remove_item_from_db (history, item);
+        gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
         g_object_unref (item);
     }
 
@@ -924,20 +850,110 @@ midori_history_popup_menu_cb (GtkWidget*     widget,
     }
 }
 
+static void
+midori_history_row_expanded_cb (GtkTreeView*   treeview,
+                                GtkTreeIter*   iter,
+                                GtkTreePath*   path,
+                                MidoriHistory* history)
+{
+    GtkTreeModel* model;
+    KatzeItem* item;
+
+    model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
+    gtk_tree_model_get (model, iter, 0, &item, -1);
+    midori_history_read_from_db (history, GTK_TREE_STORE (model),
+                                 iter, katze_item_get_added (item), NULL);
+    g_object_unref (item);
+}
+
+static void
+midori_history_row_collapsed_cb (GtkTreeView *treeview,
+                                 GtkTreeIter *parent,
+                                 GtkTreePath *path,
+                                 gpointer     user_data)
+{
+    GtkTreeModel* model;
+    GtkTreeStore* treestore;
+    GtkTreeIter child;
+
+    model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
+    treestore = GTK_TREE_STORE (model);
+    while (gtk_tree_model_iter_nth_child (model, &child, parent, 0))
+        gtk_tree_store_remove (treestore, &child);
+    /* That's an invisible dummy, so we always have an expander */
+    gtk_tree_store_insert_with_values (treestore, &child, parent,
+        0, 0, NULL, 1, NULL, -1);
+}
+
+static gboolean
+midori_history_filter_timeout_cb (gpointer data)
+{
+    MidoriHistory* history = data;
+    GtkTreeModel* model;
+    GtkTreeStore* treestore;
+
+    model = gtk_tree_view_get_model (GTK_TREE_VIEW (history->treeview));
+    treestore = GTK_TREE_STORE (model);
+
+    gtk_tree_store_clear (treestore);
+    midori_history_read_from_db (history, treestore, NULL, 0, history->filter);
+
+    return FALSE;
+}
+
+static void
+midori_history_filter_entry_changed_cb (GtkEntry*      entry,
+                                        MidoriHistory* history)
+{
+    if (history->filter_timeout)
+        g_source_remove (history->filter_timeout);
+    history->filter_timeout = g_timeout_add (COMPLETION_DELAY,
+        midori_history_filter_timeout_cb, history);
+    katze_assign (history->filter, g_strdup (gtk_entry_get_text (entry)));
+}
+#endif
+static void
+midori_history_filter_entry_clear_cb (GtkEntry*      entry,
+                                      gint           icon_pos,
+                                      gint           button,
+                                      MidoriHistory* history)
+{
+    if (icon_pos == GTK_ICON_ENTRY_SECONDARY)
+        gtk_entry_set_text (entry, "");
+}
+
+
 static void
 midori_history_init (MidoriHistory* history)
 {
+    GtkWidget* entry;
+    GtkWidget* box;
     GtkTreeStore* model;
     GtkWidget* treeview;
     GtkTreeViewColumn* column;
     GtkCellRenderer* renderer_pixbuf;
     GtkCellRenderer* renderer_text;
 
-    history->net = katze_net_new ();
-    /* FIXME: Dereference the net on finalization */
+    /* Create the filter entry */
+    entry = gtk_icon_entry_new ();
+    gtk_icon_entry_set_icon_from_stock (GTK_ICON_ENTRY (entry),
+        GTK_ICON_ENTRY_SECONDARY, GTK_STOCK_CLEAR);
+    gtk_icon_entry_set_icon_highlight (GTK_ICON_ENTRY (entry),
+        GTK_ICON_ENTRY_SECONDARY, TRUE);
+    g_signal_connect (entry, "icon-release",
+        G_CALLBACK (midori_history_filter_entry_clear_cb), history);
+    #if HAVE_SQLITE
+    g_signal_connect (entry, "changed",
+        G_CALLBACK (midori_history_filter_entry_changed_cb), history);
+    #endif
+    box = gtk_hbox_new (FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (box), gtk_label_new (_("Filter:")), FALSE, FALSE, 3);
+    gtk_box_pack_start (GTK_BOX (box), entry, TRUE, TRUE, 3);
+    gtk_widget_show_all (box);
+    gtk_box_pack_start (GTK_BOX (history), box, FALSE, FALSE, 5);
 
     /* Create the treeview */
-    model = gtk_tree_store_new (2, KATZE_TYPE_ITEM, G_TYPE_INT64);
+    model = gtk_tree_store_new (2, KATZE_TYPE_ITEM, G_TYPE_STRING);
     treeview = gtk_tree_view_new_with_model (GTK_TREE_MODEL (model));
     gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (treeview), FALSE);
     column = gtk_tree_view_column_new ();
@@ -948,11 +964,11 @@ midori_history_init (MidoriHistory* history)
         treeview, NULL);
     renderer_text = gtk_cell_renderer_text_new ();
     gtk_tree_view_column_pack_start (column, renderer_text, FALSE);
-    gtk_tree_view_column_set_cell_data_func (column, renderer_text,
-        (GtkTreeCellDataFunc)midori_history_treeview_render_text_cb,
-        history, NULL);
+    gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (column), renderer_text,
+        "text", 1, NULL);
     gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
     g_object_unref (model);
+    #if HAVE_SQLITE
     g_object_connect (treeview,
                       "signal::row-activated",
                       midori_history_row_activated_cb, history,
@@ -964,12 +980,20 @@ midori_history_init (MidoriHistory* history)
                       midori_history_button_release_event_cb, history,
                       "signal::key-release-event",
                       midori_history_key_release_event_cb, history,
+                      "signal::row-expanded",
+                      midori_history_row_expanded_cb, history,
+                      "signal::row-collapsed",
+                      midori_history_row_collapsed_cb, history,
                       "signal::popup-menu",
                       midori_history_popup_menu_cb, history,
                       NULL);
+    #endif
     gtk_widget_show (treeview);
     gtk_box_pack_start (GTK_BOX (history), treeview, TRUE, TRUE, 0);
     history->treeview = treeview;
+    /* FIXME: We need to connect a signal here, to add new pages into history */
+
+    history->filter = NULL;
 }
 
 static void
@@ -982,8 +1006,8 @@ midori_history_finalize (GObject* object)
 
     /* FIXME: We don't unref items (last argument is FALSE) because
        our reference counting is incorrect. */
-    midori_history_disconnect_folder (history, history->array, FALSE);
     g_object_unref (history->array);
+    katze_assign (history->filter, NULL);
 }
 
 /**
